@@ -12,6 +12,7 @@ import networkx as nx
 import numpy as np
 from ..core.singleton import Singleton
 import warnings
+import matplotlib.pyplot as plt
 
 class MarketMatching(metaclass=ABCMeta):
     """
@@ -148,7 +149,9 @@ class MarketMatching(metaclass=ABCMeta):
         :param agent: reference agent
         :return: list of agents
         """
-        return list(self.graph.pred[agent].keys())
+        if agent in self.graph.pred:
+            return list(self.graph.pred[agent].keys())
+        return []
 
     def get_demanders_from(self,agent) -> list:
         """
@@ -157,7 +160,9 @@ class MarketMatching(metaclass=ABCMeta):
         :param agent: reference agent
         :return: list of agents
         """
-        return list(self.graph[agent].keys())
+        if agent in self.graph:
+            return list(self.graph[agent].keys())
+        return []
 
     def link_agents(self,supply_agent,demand_agent,val):
         """
@@ -237,15 +242,21 @@ class MarketMatching(metaclass=ABCMeta):
             plt.close()
 
 
-    def plot_weighted(self,my_layout=None,arrows=False):
+    def plot_weighted(self,my_layout=None,arrows=False,fname=None):
         """
         plot the current matching situation as a network graph, with edge thickness matching the weights
         Suppliers re plotted in red, demanders in blue.
 
         :param my_layout: networkx graph layout
         :param arrows: show the arrows
+        :param fname: if None, plot is shown in window. Else, file is stored to fname
         """
         widths = nx.get_edge_attributes(self.graph, 'weight')
+        m = max(widths.values()) # max linewidth is 3.5
+
+        for k,v in widths.items():
+            widths[k] = widths[k]/m * 3.5
+        
         nodelist = self.graph.nodes()
 
         colors = []
@@ -277,4 +288,9 @@ class MarketMatching(metaclass=ABCMeta):
                        alpha=1.0,
                        arrows=arrows)
         plt.box(False)
-        plt.show()
+
+        if fname is None:
+            plt.show()
+        else:
+            plt.savefig(fname)
+            plt.close()
