@@ -67,8 +67,8 @@ class FlowMatrix(Singleton):
 
         self.initialized = True # initialized flag
 
-        # TODO replace above code with 'if self.do_init:' check?
-        # TODO Keeping it for now because it works...
+        # TODO maybe think about replacing above code with 'if self.do_init:' check?
+        # Keeping it for now because it works...
 
         self._flow_data = {}  # has two keys CA and KA, see reset()...
         self.reset()  # defaultdict will take care of missing keys here
@@ -137,10 +137,7 @@ class FlowMatrix(Singleton):
 
     def check_consistency(self):
         """
-        Checks the consistency of the flow matrix (i.e.)
-
-        NOTE is slow because of pandas dataframe usage. Do not call often, do not call in excessive for-loops
-        TODO test this
+        Checks the consistency of the transaction flow matrix.
         """
 
         df = self.to_dataframe(group=True)
@@ -247,12 +244,24 @@ class FlowMatrix(Singleton):
             df2 = df2.round(5)
             df2 = df2.round(4) # cut one digit to obtain a 'consistently rounded table'
 
-
             if insert_nullsym:
                 null_sym = "   .-   "
                 df2 = df2.replace(0.0,null_sym)
 
-            df2 = df2.reindex(sorted(df2.columns), axis=1)
+            sorted_cols2 = sorted(df2.columns)
+            try:
+                df2 = df2.reindex(sorted_cols2, axis=1)
+
+            except ValueError: # non-unique multi-indices here? bad sign... try to remove the non-unique part
+                try:
+                    sorted_cols2 = sorted(df2.columns)
+                    sorted_cols2 = list(set(sorted_cols2)) # try to remove non-unique multi-indices
+                    df2 = df2.reindex(sorted_cols2, axis=1)
+
+                    warnings.warn("Encountered non-unique multi-index in FlowMatrix. Something might be wrong")
+
+                except: # other error?
+                    raise RuntimeError("Something went wrong when aggregating the FlowMatrix as dataframe. You might be able to fix this error by cross-checking the naming of your agents and transactions.")
 
             cols = list(df2.columns.values) #Make a list of all of the columns in the df
             #print("COLS",cols)
