@@ -1,5 +1,5 @@
 
-__version__ = "0.3"
+__version__ = "0.9"
 __author__ = "Thomas Baldauf"
 __email__ = "thomas.baldauf@dlr.de"
 __license__ = "MIT"
@@ -7,7 +7,6 @@ __birthdate__ = '15.11.2021'
 __status__ = 'prod' # options are: dev, test, prod
 
 
-
 from ..core.settings import Settings
 from ..bottomup.stock_manager import StockManager
 from ..core.flow_matrix import FlowMatrix
@@ -17,11 +16,9 @@ import warnings
 import numpy as np
 from collections import defaultdict
 from enum import Enum
-from collections import defaultdict
 import copy
 
 
-
 class BalanceEntry(Enum):
     """
     these are the different entry types
@@ -32,6 +29,15 @@ class BalanceEntry(Enum):
     TOTAL = 3
 
 
+class BankruptcyEvent(Enum):
+    """
+    bankruptcy cases
+    """
+    NEGATIVE_ASSETS = 0
+    NEGATIVE_LIABILITIES = 1
+    NEGATIVE_EQUITY = 2
+
+
 class BalanceSheet:
     """
     This is the balance sheet of an agent (it could be used as a standalone datastructure, too, in theory). A balance sheet is of the following form
@@ -55,6 +61,23 @@ class BalanceSheet:
 
     """
 
+    _tolerance = 1e-6 # tolerance for engage()
+    _tolerance_rel = 1e-4 # relative tolerance of deviation from total assets
+
+    @classmethod
+    def change_tolerance(cls, tol:float, rel_tol:float):
+        """
+        change the tolerance by which the balace is allowed to diverge
+        This will set the new values globally.
+
+        :param tol: new absolute tolerance
+        :param rel_tol: relative tolerance (relative devaition from total assets)
+        """
+
+        assert tol > 0
+        cls._tolerance = tol
+        cls._tolerance_rel = rel_tol
+
     class Proxy:
         """
         Proxy class for balance sheet. It can be either engaged (consistent and 'active') or disengaged (inconsistent and 'disabled').
@@ -67,6 +90,7 @@ class BalanceSheet:
             self.parent._engaged = False # disengage()
 
         def __exit__(self, exc_type, exc_val, exc_tb):
+            self.parent.engage()
             self.parent._engaged = True # engage()
 
     def __init__(self, owner):
@@ -75,9 +99,7 @@ class BalanceSheet:
 
         :param owner: the owner agent of this balance sheet. Can be None for standalone purposes, but it is recommended to pass an Agent instance here.
         """
-
         self.owner = owner # stores owner instance, if any
-
         default_value = 0.0 # np.nan NOTE if nan is set this can lead to misbehavior in engage.
         self._sheet = defaultdict(lambda: {BalanceEntry.ASSETS: default_value,
                                           BalanceEntry.EQUITY: default_value,
@@ -89,6 +111,30 @@ class BalanceSheet:
 
         self._engaged = True  # <- this can deactivate the consistency check temporarily for modification. Manual modification is not allowed (therefore private).
         self._bankrupt = False # <- this stores if the owner went bankrupt. manual modification should not be needed (therefore private).
+        # self.tolerance -> self.__class__.tolerance # numerical tolerance for consistency cross-check
+
+        # link agents to specific entries in the balance sheet
+        # nested by [BalanceEntry]->[ItemName]->[Agent]->[float Balance]
+        self._changelog = defaultdict(lambda: defaultdict(lambda: defaultdict(lambda: 0.0)))
+
+    def add_changelog(self, agent, name, which, x):
+        """
+        inserts a link between a specific balance sheet item to a
+        for example, a bank might want to record which agent its deposits belong to...
+        Data structure: [BalanceEntry]->[ItemName]->[Agent]->[float Balance]
+
+        :param agent: agent instance to store here
+        :param name: name of the item, e.g. 'Deposits'
+        :param which: corresponding BalanceEntry
+        :param x: quantity which is inserted, can be positive, zero or negative
+        """
+        # assert isintance(x,int) or isinstance(x,float), "wrong data type"
+        y = self._changelog[which][name][agent]
+
+        if np.isnan(y):
+            self._changelog[which][name][agent] = x  # first time this is called? overwrite nan
+        else:
+            self._changelog[which][name][agent] = x + y # after first time
 
     @property
     def raw_data(self):
@@ -100,6 +146,20 @@ class BalanceSheet:
         del d[BalanceEntry.TOTAL]
         return d
 
+    def get_last_changelog(self, name, which,agent=None):
+        """
+        returns the internally stored links to other agents made via balance sheet transactions
+
+        :param agent: agent instance to store here. if None, all agents' entries will be returned
+        :param name: name of the item, e.g. 'Deposits'
+        :param which: corresponding BalanceEntry
+        :return: float, current balance or dict of balances for agents
+        """
+
+        if agent is not None:
+            return self._changelog[which][name][agent]
+        return self._changelog[which][name]
+
     #def items(self):
     #    return self.raw_data.items()
 
@@ -151,11 +211,12 @@ class BalanceSheet:
         self._engaged = False
 
     def engage(self):
-        """locks the balance sheet -> no longer can be modified
-         performs also a cross-check if consistency is maintained
-         """
+        """
+        locks the balance sheet -> no longer can be modified
+        performs also a cross-check if consistency is maintained
+        """
 
-        eps = 1e-6 # tolerance level for cross-check
+        eps = self.__class__._tolerance # tolerance level for cross-check, 1e-6 by default
 
         a = 0.0 # dummy for sum of entries
         e = 0.0
@@ -175,21 +236,32 @@ class BalanceSheet:
             e += e_i
             l += l_i
 
-            if a_i < -eps or l_i < -eps:
+            if a_i < -eps: # handle negative assets
 
-                # raise RuntimeError("...")
-
-                if True: #  entry == "Cash":
+                if True: #  TODO might add exceptions for some special cases?
                     if not self._bankrupt:
                         self._bankrupt = True
-
                         warnings.warn("filed bankruptcy for %s\n" %self.owner)
-                        # print(self.owner.balance_sheet.to_string())
-
-                        self.owner.file_bankruptcy(event="negative assets")
+                        self.owner.file_bankruptcy(event=BankruptcyEvent.NEGATIVE_ASSETS)
 
                     return None
-                else:
+
+                else: # NOTE not used as of now...
+                    s = self.to_string()
+                    error = ValueError("%s of %s reached a value below zero after engage() call" % (entry, self.owner) + "\n\n" + s)
+                    raise error
+
+            if l_i < -eps: # handle negative liabilities
+
+                if True: #  TODO might add exceptions for some special cases?
+                    if not self._bankrupt:
+                        self._bankrupt = True
+                        warnings.warn("filed bankruptcy for %s\n" %self.owner)
+                        self.owner.file_bankruptcy(event=BankruptcyEvent.NEGATIVE_LIABILITIES)
+
+                    return None
+
+                else: # NOTE not used as of now...
                     s = self.to_string()
                     error = ValueError("%s of %s reached a value below zero after engage() call" % (entry, self.owner) + "\n\n" + s)
                     raise error
@@ -197,16 +269,15 @@ class BalanceSheet:
         if abs(a - (e+l)) > eps: # check bookkeeping equation for consistency within tolerance level eps
 
             # almost zero? bookkeeping equation obeyed
-            if a !=0 and abs(a - (e+l))/abs(a) < eps:
+            eps2 = self.__class__._tolerance_rel # relative tolerance
+            if a !=0 and abs(a - (e+l))/abs(a) < eps2:
 
                 pass
-                # s = ""
-
                 # relative deviation is still small?  -> just warn
                 # warnings.warn("balance sheet of %s is numerically instable. Deviation %.10f" % (self.owner, a-(e+l)) + "\n\n" + s)
+                # NOTE ^ above code was commented out because no case could be thought of where this would make sense.
 
             else: # not almost zero? bookkeping equation violated
-
                 s = self.to_string()
                 error = ValueError("balance sheet of %s is corrupted after cross-check. Deviation %.10f" % (self.owner, a-(e+l)) + "\n\n" + s)
                 raise error
@@ -215,15 +286,10 @@ class BalanceSheet:
 
             if not self._bankrupt:
                 self._bankrupt = True
-
-                # warnings.warn("file bankruptcy for %s\n" %self.owner)
-
-                self.owner.file_bankruptcy(event="negative equity")
-
+                self.owner.file_bankruptcy(event=BankruptcyEvent.NEGATIVE_EQUITY)
             else:
                 # already bankrupt -> do not trigger bankruptcy a second time... # TODO < think about logic in here
                 pass
-
             return None
 
     def get_balance(self, key, kind=BalanceEntry.ASSETS) -> float:
@@ -241,21 +307,26 @@ class BalanceSheet:
         if isinstance(kind, str):
             mydict = {
                 "Assets": BalanceEntry.ASSETS,
-                BalanceEntry.ASSETS: BalanceEntry.ASSETS,
+                # BalanceEntry.ASSETS: BalanceEntry.ASSETS,
                 "Equity": BalanceEntry.EQUITY,
-                BalanceEntry.EQUITY: BalanceEntry.EQUITY,
+                # BalanceEntry.EQUITY: BalanceEntry.EQUITY,
                 "Liabilities": BalanceEntry.LIABILITIES,
-                BalanceEntry.LIABILITIES: BalanceEntry.LIABILITIES,
+                # BalanceEntry.LIABILITIES: BalanceEntry.LIABILITIES,
             }
             which = mydict[kind]
         else:
-            which = kind
+            which = kind # should be BalanceEntry at this point! Will raise KeyError if not
+
+        if not which in self._sheet[key]:
+            raise KeyError("Could not find entry for %s"% which)
 
         return_value = self._sheet[key][which]
-        if np.isnan(return_value):
-            return 0
-        else:
-            return return_value
+
+        if np.isnan(return_value): # nan values are not allowed, so throw an error
+            msg = "Requested Value is NaN (not a number)."
+            raise RuntimeError(msg)
+
+        return return_value
 
     def to_string(self):
         """
@@ -263,6 +334,7 @@ class BalanceSheet:
 
         :return: str, balance sheet string representation
         """
+        # TODO think of different graphical representations, e.g. entries in rows instead of columns
         return "\n\n" + self.to_dataframe().to_string().replace("NaN"," ~ ") + "\n\n"
 
     def change_item(self, name, which, value, suppress_stock=False):
@@ -274,37 +346,27 @@ class BalanceSheet:
         :param which: BalanceEntry: Asset, Liability, Equity
         """
 
-        """
-        if isinstance(which, str):
-            mydict = {
-                "Assets": BalanceEntry.ASSETS,
-                BalanceEntry.ASSETS: BalanceEntry.ASSETS,
-                "Equity": BalanceEntry.EQUITY,
-                BalanceEntry.EQUITY: BalanceEntry.EQUITY,
-                "Liabilities": BalanceEntry.LIABILITIES,
-                BalanceEntry.LIABILITIES: BalanceEntry.LIABILITIES,
-            }
-
-            which = mydict[which]
-        """
-
         if value == 0:  # security lock if this is a 'ghost transaction'
+            warnings.warn("Blocked ghost item in balance (value=0): %s, %s,%s,%s"% (self,name,which,value))
             return
 
+        if self._engaged: # security lock when balance is still disengaged
+            raise PermissionError("Cannot change item in engaged balance sheet! This would lead to inconsistencies. Please disengage before calling change_item or use 'modify' proxy.")
 
-        if self._engaged:
-            raise PermissionError("Cannot change item in engaged balance sheet! This would lead to inconsistencies. Please disengage before calling change_item")
+        if np.isnan(self._sheet[name][which]):
+            warnings.warn("Found NaN value in %s, %s, %s" % (self,name,which))
+            # self._sheet[name][which] = 0 # NOTE <- alternative way to handle
 
-        if np.isnan(self._sheet[name][which]): # first time initialized? -> change nan to zero
-            self._sheet[name][which] = 0
-
-        if np.isnan(self[BalanceEntry.TOTAL][which]): # first time initialized? -> change nan to zero
-            self._sheet[BalanceEntry.TOTAL][which] = 0
+        if np.isnan(self[BalanceEntry.TOTAL][which]):
+            warnings.warn("Found NaN value in %s, %s, %s" % (self,"TOTAL",which))
+            # self._sheet[BalanceEntry.TOTAL][which] = 0 # NOTE <- alternative way to handle
 
         self._sheet[name][which] += value
         self._sheet[BalanceEntry.TOTAL][which] += value
 
         if not suppress_stock:
+            # Suppresses logging in the transaction flow matrix. This can be helpful in some cases
+
             if which == BalanceEntry.ASSETS:
                 FlowMatrix().capital_flow_data["Δ %s" % name][self.owner] += -value #-value
             if which == BalanceEntry.LIABILITIES:
@@ -313,9 +375,9 @@ class BalanceSheet:
             if which == BalanceEntry.EQUITY and name == "Cash":
                 raise ValueError("Cash cannot be equity")
 
+
     """
-    TODO insert operations like 'Aktivtausch', 'Passivtausch', 'Bilanzverlängerung'(?)
-    TODO < insert code here
+    TODO (?) insert operations like 'Aktivtausch', 'Passivtausch', 'Bilanzverlängerung'(?)
     """
 
     def to_dataframe(self):
@@ -325,8 +387,8 @@ class BalanceSheet:
         :return: pandas dataframe
         """
 
-        df = pd.DataFrame({"Commodity": [],"Assets": [], "Equity": [],"Liabilities": []}) # construct an empty 'dummy' data frame
-        coms = [] # list of commodities
+        df = pd.DataFrame({"Entry": [],"Assets": [], "Equity": [],"Liabilities": []}) # construct an empty 'dummy' data frame
+        entries = [] # allocate list of commodities entered in the balance sheet
 
         null_sym = "   .-   " # symbol for zero entries
 
@@ -334,30 +396,26 @@ class BalanceSheet:
         for key, row in self._sheet.items():
 
             if key != BalanceEntry.TOTAL: # consider all rows except the 'total' entry
-
-                coms.append(key) # commodities
-
-                new_row = pd.DataFrame({"Commodity": [key],
+                entries.append(key) # add to entries list
+                new_row = pd.DataFrame({"Entry": [key],
                                         "Assets": [row[BalanceEntry.ASSETS]],
                                         "Equity": [row[BalanceEntry.EQUITY]],
                                         "Liabilities": [row[BalanceEntry.LIABILITIES]]}) # construct a new row
-
-                # df = df.append(new_row.replace(0.0,null_sym)) # append the new row
                 df =  pd.concat([df,new_row.replace(0.0,null_sym)])
 
         df = df.sort_index() # sort the entries by index
 
-        prices = [1.0 for key in coms] # dummy for prices. NOTE TODO Might change with exchange rates or any other conversion factors (?)
-        # for now just assume all values are anyway nominal and therefore no adjustment is required.
+        # Compute 'Total' row
+        # dummy for modifying factors.
+        # NOTE is 1 always for now, but might change with exchange rates or any other conversion factors (?)
+        factors = [1.0 for key in entries]
+        my_row = pd.DataFrame({"Entry": ["Total"],
+                                "Assets": [np.dot(df["Assets"].replace(null_sym,0.0),factors)],
+                                "Equity": [np.dot(df["Equity"].replace(null_sym,0.0),factors)],
+                                "Liabilities": [np.dot(df["Liabilities"].replace(null_sym,0.0),factors)]})
+        df = pd.concat([df,my_row]) # adds the 'Total' row
 
-        nominal_row = pd.DataFrame({"Commodity": ["Total"],
-                                "Assets": [np.dot(df["Assets"].replace(null_sym,0.0),prices)],
-                                "Equity": [np.dot(df["Equity"].replace(null_sym,0.0),prices)],
-                                "Liabilities": [np.dot(df["Liabilities"].replace(null_sym,0.0),prices)]})
-        # df = df.append(nominal_row)
-        df = pd.concat([df,nominal_row])
-
-        df = df.set_index("Commodity") # sets index column correctly
+        df = df.set_index("Entry") # sets index column correctly
         df.index.name = "BALANCE SHEET OF %s" % self.owner # gives the balance sheet an understandable title
 
         return df
@@ -369,14 +427,15 @@ class BalanceSheet:
         Equity will become less, liabilities will stay.
         """
 
-        settings = Settings()
-        # read depreciation rates from settings
+        settings = Settings() # read depreciation rates from settings
+
         for item_name in self._sheet.keys():
 
             depr_rate = settings.config_data["params"][item_name]["depreciation"]
-            depr_q = self._sheet[item_name][BalanceEntry.EQUITY] * depr_rate
+            # Settings entry is 'param->(item name)->depreciation->value'
 
-            warnings.warn("Depreciation is percentage of current stock. Non-linear!")
+            depr_q = self._sheet[item_name][BalanceEntry.EQUITY] * depr_rate
+            # Warning: Depreciation is percentage of current stock. Non-linear!
 
             if not np.isnan(self._sheet[item_name][BalanceEntry.EQUITY]):
                 self._sheet[item_name][BalanceEntry.EQUITY] -= depr_q
@@ -384,6 +443,8 @@ class BalanceSheet:
             if not np.isnan(self._sheet[item_name][BalanceEntry.ASSETS]):
                 self._sheet[item_name][BalanceEntry.ASSETS] -= depr_q
 
+        # TODO maybe automatically log this on income statement of the underlying agent here?
+
     @property
     def leverage(self) -> float:
         """
@@ -460,11 +521,13 @@ class BalanceSheet:
         blues.reverse()
         reds.reverse()
 
+        # color patches for legend
         legend_patches = []
         blue_patches = []
         green_patches = []
         red_patches = []
 
+        # data for plotting
         data = {"Assets": [], "Liabilities & Equity": []}
         labels = {}
         i = 0
@@ -473,6 +536,7 @@ class BalanceSheet:
         lia = []
         equi = []
 
+        # construct data for plotting
         for key, val in balance_sheet._sheet.items():
 
             if key != BalanceEntry.TOTAL:
@@ -500,6 +564,7 @@ class BalanceSheet:
                     assets.append([key, val[BalanceEntry.ASSETS]])
                     lia.append([key, 0.0])
 
+        # construct data for plotting
         for a in assets:
 
             key = a[0]
@@ -547,6 +612,7 @@ class BalanceSheet:
                 labels[i] = key
                 i += 1
 
+        # create customized patche for the legend
         legend_patches = red_patches + blue_patches + green_patches
 
         df = pd.DataFrame(data).T
