@@ -28,21 +28,21 @@ class BankOrderBook:
 
     Loans:
 
-    +----------+-----------+----------+-------------------------+---------------------------------+------+----------------+
-    |  ID      | Agent     | Loans    | interest on loans (r_l) |   Loan Payback rate (absolute)  | Time | Loan Duration  |
-    +----------+-----------+----------+-------------------------+---------------------------------+------+----------------+
-    |     0    |  10.0     |          | 0.05                    |  0.01                           | ...  |                |
-    +----------+-----------+----------+-------------------------+---------------------------------+------+----------------+
-    |     1    |   ...     |          |                         |                                 |      |                |
-    +----------+-----------+----------+-------------------------+---------------------------------+------+----------------+
-    |          |           |          |                         |                                 |      |                |
-    +----------+-----------+----------+-------------------------+---------------------------------+------+----------------+
-    |          |   ...     |          |                         |                                 |      |                |
-    +----------+-----------+----------+-------------------------+---------------------------------+------+----------------+
-    |          |           |          |                         |                                 |      |                |
-    +----------+-----------+----------+-------------------------+---------------------------------+------+----------------+
-    |     42   |   ...     |          |                         |                                 |      |                |
-    +----------+-----------+----------+-------------------------+---------------------------------+------+----------------+
+    +----------+-----------+-------------------+-------------------------+---------------------------------+------+----------------+
+    |  ID      | Agent     | Outstanding Loans | interest on loans (r_l) |  Principal                      | Time | Maturity       |
+    +----------+-----------+-------------------+-------------------------+---------------------------------+------+----------------+
+    |     0    |  10.0     | 10.0              | 0.05                    |  100.0                          | ...  |                |
+    +----------+-----------+-------------------+-------------------------+---------------------------------+------+----------------+
+    |     1    |   ...     |                   |                         |                                 |      |                |
+    +----------+-----------+-------------------+-------------------------+---------------------------------+------+----------------+
+    |          |           |                   |                         |                                 |      |                |
+    +----------+-----------+-------------------+-------------------------+---------------------------------+------+----------------+
+    |          |   ...     |                   |                         |                                 |      |                |
+    +----------+-----------+-------------------+-------------------------+---------------------------------+------+----------------+
+    |          |           |                   |                         |                                 |      |                |
+    +----------+-----------+-------------------+-------------------------+---------------------------------+------+----------------+
+    |     42   |   ...     |                   |                         |                                 |      |                |
+    +----------+-----------+-------------------+-------------------------+---------------------------------+------+----------------+
 
 
     Deposits:
@@ -65,22 +65,27 @@ class BankOrderBook:
 
     """
 
-    def __init__(self, owner,credit_market:MarketMatching,deposit_market:MarketMatching):
+    def __init__(self, owner,credit_market:MarketMatching=None,deposit_market:MarketMatching=None):
         """
         constructor
 
         :param owner: agent instance who owns this data structure
-        :credit_market: a MarketMatching obect for the credit market
-        :deposit_market:  a MarketMatching object or the deposit market
+        :credit_market: a MarketMatching obect for the credit market (default None). Providing an instance will automatically unlink agents when the deposit is dissolved.
+        :deposit_market:  a MarketMatching object or the deposit market (default None). Providing an instance will automatically unlink agents when the deposit is dissolved.
 
         """
 
         self.owner = owner
         self.loans_data = defaultdict(list)
+
+        """
+        defaultdata of loans
+        """
+
         self.deposits_data = defaultdict(list)
 
         """
-        deposits, loans, r_d, r_l
+        defaultdict of deposits
         """
 
         self.idx_d = 0  # index counter deposits
@@ -102,21 +107,20 @@ class BankOrderBook:
         :return: deposits_df, loans_df (tuple)
         """
 
-        data = {"ID":[],"Agent":[],"Loans":[],"r_l":[],"payback rate":[],"time":[],"duration":[]}
+        data = {"ID":[],"Agent":[],"Outstanding Loans":[],"Principal": [], "r_l":[],"Time":[],"Maturity":[]}
         for agent, listing in self.loans_data.items():
             for entry in listing:
                 data["ID"].append(entry["idx"])
                 data["Agent"].append(agent)
-                data["Loans"].append(entry["Loans"])
+                data["Outstanding Loans"].append(entry["Outstanding Loans"])
                 data["r_l"].append(entry["r_l"])
-
-                data["payback rate"].append(entry["payback"])
-                data["time"].append(entry["time"])
-                data["duration"].append(entry["loan duration"])
+                data["Principal"].append(entry["Principal"])
+                data["Time"].append(entry["Time"])
+                data["Maturity"].append(entry["Maturity"])
 
         loans_df = pd.DataFrame(data).set_index("ID")
 
-        data = {"ID": [], "Agent":[], "Deposits": [], "r_d": [], "time":[],"duration":[]}
+        data = {"ID": [], "Agent":[], "Deposits": [], "r_d": [], "Time":[],"Duration":[]}
 
         for agent,listing in self.deposits_data.items():
             for entry in listing:
@@ -124,8 +128,8 @@ class BankOrderBook:
                 data["Agent"].append(agent)
                 data["Deposits"].append(entry["Deposits"])
                 data["r_d"].append(entry["r_d"])
-                data["time"].append(entry["time"])
-                data["duration"].append(entry["duration"])
+                data["Time"].append(entry["Time"])
+                data["Duration"].append(entry["Duration"])
 
         deposits_df =  pd.DataFrame(data).set_index("ID")
 
@@ -163,7 +167,8 @@ class BankOrderBook:
         :param agent: agent belonging to the loan
         """
 
-        self.credit_market.unlink_agents(supply_agent=self.owner, demand_agent=agent)
+        if self.credit_market is not None:
+            self.credit_market.unlink_agents(supply_agent=self.owner, demand_agent=agent)
 
         remove_index = None
         # find the entry with the corresponding index
@@ -174,7 +179,7 @@ class BankOrderBook:
         if remove_index is None:
             raise KeyError("Could not find entry")
 
-        self.total_loans -= self.loans_data[agent][remove_index]["Loans"]
+        self.total_loans -= self.loans_data[agent][remove_index]["Outstanding Loans"]
         del self.loans_data[agent][remove_index]
 
     def remove_deposit(self, idx, agent):
@@ -185,7 +190,8 @@ class BankOrderBook:
         :param agent: agent who belongs to this deposit
         """
 
-        self.deposit_market.unlink_agents(supply_agent=agent, demand_agent=self.owner)
+        if self.deposit_market is not None:
+            self.deposit_market.unlink_agents(supply_agent=agent, demand_agent=self.owner)
 
         remove_index = None
         # find the entry with the corresponding index
@@ -256,8 +262,8 @@ class BankOrderBook:
             "idx":self.idx_d,
             "Deposits": new_deposits,
             "r_d":r_d,
-            "time":0,
-            "duration":t
+            "Time":0,
+            "Duration":t
         })
 
         self.total_deposits += new_deposits
@@ -271,11 +277,11 @@ class BankOrderBook:
         for key,val in self.deposits_data.items():
 
             agent = key
-            deposits = 0.0
+            x = 0.0
             for entry in val:
-                deposits += entry["Deposits"]
+                x += entry["Deposits"]
 
-            deps[agent] = deposits
+            deps[agent] = x
 
         return deps
 
@@ -287,11 +293,11 @@ class BankOrderBook:
         for key, val in self.loans_data.items():
 
             agent = key
-            deposits = 0.0
+            x = 0.0
             for entry in val:
-                deposits += entry["Loans"]
+                x += entry["Outstanding Loans"]
 
-            loans[agent] = deposits
+            loans[agent] = x
 
         return loans
 
@@ -382,12 +388,13 @@ class BankOrderBook:
 
         self.loans_data[agent].append({
             "idx": self.idx_l,
-            "Loans": new_loans,
+            "Outstanding Loans": new_loans,
             "r_l": r_l,
-            "payback": new_loans/(1.0*t),
-            "time": 0,
-            "loan duration": t
-        })
+            # "payback": new_loans/(1.0*t), DEPRECATED TB 23/11/22
+            "Principal": new_loans,
+            "Time": 0,
+            "Maturity": t
+        }) # data = {"ID":[],"Agent":[],"Outstanding Loans":[],"Principal": [], "r_l":[],"Time":[],"Maturity":[]}
 
         self.total_loans += new_loans
 
@@ -409,18 +416,20 @@ class BankOrderBook:
             if cumu >= 0:
                 break
 
-            if entry["Loans"] <= q - cumu:  # require whole entry?
-                cumu += entry["Loans"]
+            if entry["Outstanding Loans"] <= q - cumu:  # require whole entry?
+                cumu += entry["Outstanding Loans"]
                 self.remove_loan(entry["idx"], agent)
 
             else:  # require only part of entry?
                 cumu += q - cumu
                 self.total_loans -= q - cumu
 
-                self.loans_data[agent][i]["Loans"] -= q - cumu  # partially reduce entry
+                self.loans_data[agent][i]["Outstanding Loans"] -= q - cumu  # partially reduce entry
+
                 # re-adjust the amount of loan payback
-                t_remain = self.loans_data[agent][i]["loan duration"] - self.loans_data[agent][i]["time"]
-                self.loans_data[agent][i]["payback"] = self.loans_data[agent][i]["Loans"] / (1.0 * t_remain)
+                # t_remain = self.loans_data[agent][i]["Principal"] - self.loans_data[agent][i]["time"]
+                # self.loans_data[agent][i]["payback"] = self.loans_data[agent][i]["Loans"] / (1.0 * t_remain)
+                # DEPRECATED TB 23/11/22
 
                 return
 
@@ -443,17 +452,18 @@ class BankOrderBook:
             if cumu >= q:
                 break
 
-            if entry["Loans"] <= q-cumu:  # require whole entry?
+            if entry["Outstanding Loans"] <= q-cumu:  # require whole entry?
                 self.remove_loan(entry["idx"],agent)
-                cumu += entry["Loans"]
+                cumu += entry["Outstanding Loans"]
 
             else:  # require only part of entry?
-                self.loans_data[agent][i]["Loans"] -= q-cumu  # partially reduce entry
+                self.loans_data[agent][i]["Outstanding Loans"] -= q-cumu  # partially reduce entry
                 cumu += q-cumu
                 self.total_loans -= q-cumu
 
                 # re-adjust the amount of loan payback
-                t_remain = self.loans_data[agent][i]["loan duration"] - self.loans_data[agent][i]["time"]
-                self.loans_data[agent][i]["payback"] = self.loans_data[agent][i]["Loans"] / (1.0*t_remain)
+                # t_remain = self.loans_data[agent][i]["Maturity"] - self.loans_data[agent][i]["time"]
+                # self.loans_data[agent][i]["payback"] = self.loans_data[agent][i]["Loans"] / (1.0*t_remain)
+                # DEPRECATED TB 23/11/22
 
                 return
