@@ -68,7 +68,7 @@ def add_value_labels(ax, spacing=.9, fmt="{0:d}", stacked=False, legend="best"):
 
 
 def matplotlib_barplot(data, xlabel, ylabel, title, color="indigo", size=(6, 5), tight_layout=True,
-                       fmt="{0:f}", stacked=False, show_labels=True, legend="best", show=True):
+                       fmt="{0:f}", stacked=False, show_labels=True, legend="best", show=True,yerr=None,ax=None):
     """
     Creates a bar plot in matplotlib. The 'macro' plotting style will be used if it is set up
 
@@ -83,6 +83,8 @@ def matplotlib_barplot(data, xlabel, ylabel, title, color="indigo", size=(6, 5),
     :paarm show_labels: show the column labels at bottom? (boolean)
     :param legend: location of legend, or 'off'
     :param show: if False, only the figure is returned. Else plot is shown
+    :param ax: axis to plot on
+    :param yerr: data for error bar (if any)
     """
 
     # if legend is 'off', no legend is shown
@@ -104,7 +106,9 @@ def matplotlib_barplot(data, xlabel, ylabel, title, color="indigo", size=(6, 5),
         data.set_index(ylabel)
 
     # plt.figure(figsize=size)
-    data.plot.bar(color=color, stacked=stacked,ax=plt.gca())
+    if ax is None:
+        ax = plt.gca()
+    data.plot.bar(color=color, stacked=stacked,ax=ax,yerr=yerr,capsize=6,ecolor="black")
     plt.gca().set_title(title)
     # plt.tick_params(axis='x', which='both', bottom=False, top=False, labelbottom=True)
     plt.ylabel(ylabel)
@@ -136,7 +140,7 @@ def matplotlib_barplot(data, xlabel, ylabel, title, color="indigo", size=(6, 5),
         return plt.gcf()
 
 
-def matplotlib_lineplot(data, xlabel=None, ylabel=None, title="", xlim=None, ylim=None, color="indigo", legend="best",marker=None,show=False):
+def matplotlib_lineplot(data, xlabel=None, ylabel=None, title="", xlim=None, ylim=None, color="indigo", legend="best",marker=None,show=False,ax=None):
     """
     creates a line plot in matplotlib. The 'macro' plotting style will be used if it is set up
 
@@ -147,6 +151,11 @@ def matplotlib_lineplot(data, xlabel=None, ylabel=None, title="", xlim=None, yli
     :param xlim, ylim: tuples of plot limits
     :param legend: location of legend, or 'off', or 'outside'
     :param show: if False, only the figure is returned. Else plot is shown
+    :param ax: axis to plot on
+
+    when overlapping barplot and lineplot, make sure you add an additional axis, see
+    https://stackoverflow.com/questions/42948576/pandas-plot-does-not-overlay
+    
     """
     try:
         plt.style.use("macro")
@@ -160,8 +169,12 @@ def matplotlib_lineplot(data, xlabel=None, ylabel=None, title="", xlim=None, yli
         data = pd.DataFrame({ylabel: data})
         data.set_index(ylabel)
 
+    if ax is None:
+        ax = plt.gca()
+
     for col in data.columns:
-        plt.plot(data[col],color=color,marker=marker,label=col)
+        ax.plot(data[col],color=color,marker=marker,label=col)
+
     plt.xlim(xlim)
     plt.ylim(ylim)
     plt.title(title)
